! Description:
!> @file
!!   AD of MFASIS fast visible/near-IR scattering model.
!
!> @brief
!!   AD of MFASIS fast visible/near-IR scattering model.
!!
!!
!! @param[out]    err               status on exit
!! @param[in]     chanprof          specifies channels and profiles to simulate
!! @param[in]     chanflag          flags to indicate which channels with LUT available
!! @param[in]     opts              options to configure the simulations
!! @param[in]     profiles          input atmospheric profiles and surface variables
!! @param[in]     coefs             coefficients structure for instrument to simulate
!! @param[in]     ircld             information on cloud columns
!! @param[in,out] ircld_ad          cloud column increments
!! @param[in]     aux               additional internal profile variables
!! @param[in,out] aux_ad            additional internal profile variable increments
!! @param[in]     reflectance       surface BRDFs
!! @param[in,out] reflectance_ad    surface BRDF increments
!! @param[in]     solar_spectrum    TOA solar irradiance for each channel
!! @param[in]     trans_scatt_ir    cloud/aerosol optical depths
!! @param[in,out] trans_scatt_ir_ad cloud/aerosol optical depth increments
!! @param[in]     mfasis_refl       quantities computed by rttov_mfasis used by TL/AD/K
!! @param[in,out] radiance_ad       input gradient wrt radiances
!
! Copyright:
!    This software was developed within the context of
!    the EUMETSAT Satellite Application Facility on
!    Numerical Weather Prediction (NWP SAF), under the
!    Cooperation Agreement dated 25 November 1998, between
!    EUMETSAT and the Met Office, UK, by one or more partners
!    within the NWP SAF. The partners in the NWP SAF are
!    the Met Office, ECMWF, KNMI and MeteoFrance.
!
!    Copyright 2017, EUMETSAT, All Rights Reserved.
!
SUBROUTINE rttov_mfasis_ad( &
              err,              &
              chanprof,         &
              chanflag,         &
              opts,             &
              profiles,         &
              coefs,            &
              ircld,            &
              ircld_ad,         &
              aux,              &
              aux_ad,           &
              reflectance,      &
              reflectance_ad,   &
              solar_spectrum,   &
              trans_scatt_ir,   &
              trans_scatt_ir_ad,&
              mfasis_refl,      &
              radiance_ad)

#include "throw.h"

  USE parkind1, ONLY : jpim, jprb, jplm

  USE rttov_types, ONLY :        &
    rttov_chanprof,              &
    rttov_options,               &
    rttov_profile,               &
    rttov_coefs,                 &
    rttov_ircld,                 &
    rttov_profile_aux,           &
    rttov_reflectance,           &
    rttov_transmission_scatt_ir, &
    rttov_radiance,              &
    rttov_mfasis_refl
!INTF_OFF
  USE rttov_types, ONLY : &
    rttov_coef_mfasis,           &
    rttov_mfasis_axis

  USE rttov_const, ONLY : &
    wcl_deff,                    &
    pi,                          &
    deg2rad,                     &
    pi_r,                        &
    clw_scheme_deff,             &
    mfasis_cld,                  &
!     mfasis_aer,                  &
    mfasis_dim_albedo,           &
    mfasis_dim_kfourier,         &
    mfasis_dim_lfourier,         &
    mfasis_dim_opdp,             &
    mfasis_dim_effdia,           &
    mfasis_dim_scaangle,         &
    nwcl_max
!INTF_ON
  IMPLICIT NONE
  INTEGER(jpim),                      INTENT(OUT)   :: err
  TYPE(rttov_chanprof),               INTENT(IN)    :: chanprof(:)
  LOGICAL(jplm),                      INTENT(IN)    :: chanflag(SIZE(chanprof))
  TYPE(rttov_options),                INTENT(IN)    :: opts
  TYPE(rttov_profile),                INTENT(IN)    :: profiles(:)
  TYPE(rttov_coefs),                  INTENT(IN)    :: coefs
  TYPE(rttov_ircld),                  INTENT(IN)    :: ircld
  TYPE(rttov_ircld),                  INTENT(INOUT) :: ircld_ad
  TYPE(rttov_profile_aux),            INTENT(IN)    :: aux
  TYPE(rttov_profile_aux),            INTENT(INOUT) :: aux_ad
  TYPE(rttov_reflectance),            INTENT(IN)    :: reflectance(SIZE(chanprof))
  TYPE(rttov_reflectance),            INTENT(INOUT) :: reflectance_ad(SIZE(chanprof))
  REAL(jprb),                         INTENT(IN)    :: solar_spectrum(SIZE(chanprof))
  TYPE(rttov_transmission_scatt_ir),  INTENT(IN)    :: trans_scatt_ir
  TYPE(rttov_transmission_scatt_ir),  INTENT(INOUT) :: trans_scatt_ir_ad
  TYPE(rttov_mfasis_refl),            INTENT(IN)    :: mfasis_refl(0:,:)
  TYPE(rttov_radiance),               INTENT(INOUT) :: radiance_ad
!INTF_END

  INTEGER(jpim)              :: nchanprof, prof, nstrms, npar
  INTEGER(jpim)              :: nlay, ndim, nid
  INTEGER(jpim)              :: d, n, i, j, k, cs, par
  TYPE(rttov_coef_mfasis)    :: mfasis_coefs
  REAL(jprb),    ALLOCATABLE :: ip   (:)
  REAL(jprb),    ALLOCATABLE :: ip_ad(:), iw_ad(:)
  INTEGER(jpim), ALLOCATABLE :: di(:)
  REAL(jprb),    ALLOCATABLE :: od   (:,:), od_eff   (:), ed   (:,:), ed_eff(:), ed_aux(:)
  REAL(jprb),    ALLOCATABLE :: od_v1(:,:), od_eff_v1(:), ed_eff_v1(:) !versions of nonlinear quantities needed for TL/AD/K
  REAL(jprb),    ALLOCATABLE :: od_ad(:,:), od_eff_ad(:), ed_ad(:,:), ed_eff_ad(:)

  REAL(jprb)                 :: albedo
  REAL(jprb)                 :: albedo_ad
  TYPE(rttov_mfasis_axis)    :: axis
  REAL(jprb)                 :: strwei_clr, strwei_cs
  REAL(jprb)                 :: refl_ad, rflstr_ad, refl_clr_ad, strwei_ad
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  REAL(jprb)                 :: od1_int,od1_mixed,od2_mixed
  REAL(jprb)                 :: od1_int_ad,od1_mixed_ad,od2_mixed_ad, ooo1_ad
  REAL(jprb),    ALLOCATABLE :: ooo1(:)
  REAL(jprb),    ALLOCATABLE :: fac_mx(:)
  REAL(jprb),    ALLOCATABLE :: fac_mx_ad(:)
  REAL(jprb)                 :: fac_sw,xxx1,xxx2            ! variables    used for switching on/off mixed cloud correction
  REAL(jprb)                 :: fac_sw_ad,xxx1_ad,xxx2_ad   ! variables    used for switching on/off mixed cloud correction
  ! --------------------------------------------------------------------------------------

  TRY

  ! TODO: Adapt to water vapor correction using three LUTs
  ! TODO: Test aerosol functionality

  ! --------------------------------------------------------------------------
  ! Initialisation
  ! --------------------------------------------------------------------------
  nchanprof = SIZE(chanprof)
  nlay = profiles(1)%nlayers

  ! Check whether cloud or aerosol simulation (currently both simultaneously not supported)
  IF (opts%rt_ir%addclouds ) THEN
    mfasis_coefs = coefs%coef_mfasis_cld
  ELSE
    mfasis_coefs = coefs%coef_mfasis_aer
  ENDIF

  npar = mfasis_coefs%nparticles        ! # of particles: 2 for clouds (water and ice)
                                        !                 or # of aerosols considered
  ndim = mfasis_coefs%ndims             ! # of LUT dimensions

  ! Find number of dimensions that are not interpolated (fourier coefs and albedo)
  n = 0_jpim
  DO d = 1, ndim
    IF (mfasis_coefs%lut_axes(d)%dim_type /= mfasis_dim_kfourier .AND. &
        mfasis_coefs%lut_axes(d)%dim_type /= mfasis_dim_lfourier .AND. &
        mfasis_coefs%lut_axes(d)%dim_type /= mfasis_dim_albedo ) CYCLE
    n = n + 1
  END DO

  nid = ndim - n ! Total number of dimensions to be interpolated (total - albedo - fourier indices)

  ALLOCATE( ip     (nid) , &      ! interpolation point
            ip_ad  (nid) , &      ! interpolation point
            iw_ad  (nid) , &      ! interpolation weight
            di     (nid), STAT=err )  ! original dimension index in nid arrays
  THROWM(err.NE.0,"Allocation of memory for rttov_mfasis_ad failed")


  ! Populate di and set some relevant dimensions indices
  di(:) = -1_jpim ! Indices of dimensions to interpolate

  n = 1_jpim
  DO d = 1, ndim
    SELECT CASE (mfasis_coefs%lut_axes(d)%dim_type)
    CASE (mfasis_dim_kfourier, mfasis_dim_lfourier, mfasis_dim_albedo)

    CASE DEFAULT
      di(n) = d
      n = n + 1
    END SELECT
  ENDDO


  ALLOCATE( od_eff    (npar)    , & ! Effective total optical depth per type of particle
            od_eff_v1 (npar)    , & ! version of od_eff for adjoint computations
            od_eff_ad (npar)    , & ! Effective total optical depth per type of particle
            od     (npar, nlay) , & ! Effective optical depth per type of particle and layer
            od_v1  (npar, nlay), &     ! od saved for adjoint computations
            od_ad  (npar, nlay), STAT=err)  ! Effective optical depth per type of particle and layer
  THROWM(err.NE.0,"Allocation of memory for rttov_mfasis_ad failed")

  ! For the time being we assume effective diameters will be out of LUT for aerosols
  IF ( mfasis_coefs%file_type .EQ. 1 ) THEN
    ALLOCATE( ed_eff (npar)      , & ! Total effective diameter per type of particle
              ed_eff_v1(npar) ,  &      ! version of ed_eff for adjoint computations
              ed_aux (npar)      , & ! Effective diameters to be used with 0 opdp to avoid flagging
              ed_eff_ad(npar)    , & ! Total effective diameter per type of particle
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
              fac_mx (nlay),      &         ! fraction of cloud considered as (potentially) mixed cloud
              fac_mx_ad (nlay),   &         ! fraction of cloud considered as (potentially) mixed cloud
              ooo1 (nlay),        &         ! discriminator for switching on mixed layer summation
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
              ed   (npar, nlay)  , & ! Effective diameter per type of particle and layer
              ed_ad(npar, nlay), STAT=err) ! Effective diameter per type of particle and layer
    THROWM(err.NE.0,"Allocation of memory for rttov_mfasis_ad failed")

    ! Effective diameters to be used with 0 opdp to avoid incorrect flagging for clear stream
    n = 1_jpim
    DO d = 1, ndim
      IF (mfasis_coefs%lut_axes(d)%dim_type .NE. mfasis_dim_effdia) CYCLE
      ed_aux(n) = mfasis_coefs%lut_axes(d)%values(1)
      n = n + 1
    ENDDO
  ENDIF
  fac_mx (:) = 0
  ooo1 (:)   = 0

!=====================================================================================
! Make sure all lokal linear variables are initialized as zeros
!=====================================================================================
  ip_ad(:)   = 0
  iw_ad(:)   = 0

  od_ad     (:,:)  = 0
  od_eff_ad (:)    = 0
  ed_ad     (:,:)  = 0
  ed_eff_ad (:)    = 0
  fac_mx_ad (:)    = 0

  albedo_ad  = 0
  refl_ad    = 0
  rflstr_ad    = 0
  refl_clr_ad  = 0
  strwei_ad    = 0

  od1_int_ad    = 0
  od1_mixed_ad  = 0
  od2_mixed_ad  = 0
  ooo1_ad       = 0

  fac_sw_ad   = 0
  xxx1_ad   = 0
  xxx2_ad   = 0
!=====================================================================================

  ! --------------------------------------------------------------------------
  ! Channel loop
  ! --------------------------------------------------------------------------
  DO i = 1, nchanprof
    IF (.NOT. chanflag(i)) CYCLE
    prof = chanprof(i)%prof

    strwei_clr = ircld%xstrclr(prof) ! clear stream

    ip(:) = -1

    IF ( mfasis_coefs%file_type == mfasis_cld ) THEN
      nstrms = ircld%nstream(prof)
    ELSE
      nstrms = 1._jpim
    ENDIF


!=====================================================================================
!  linear computations independent of stream cs
!=====================================================================================
!================================
!    radiance_ad%total(i) = refl_ad * solar_spectrum(i) * pi_r  &
!      * COS(profiles(prof)%sunzenangle * deg2rad)

! To be consistent with other RTTOV calls we only use input AD/K increments in
! radiance%total and we assume all other radiance arrays are zero
    refl_ad    = refl_ad     + solar_spectrum(i) * pi_r   &
      * COS(profiles(prof)%sunzenangle * deg2rad) * radiance_ad%total(i)



    strwei_ad  =strwei_ad   + mfasis_refl(0,i)%refl*refl_ad
    refl_clr_ad=refl_clr_ad + strwei_clr     *refl_ad

!    strwei_ad = ircld_ad%xstrclr(prof) ! clear stream
    ircld_ad%xstrclr(prof)=ircld_ad%xstrclr(prof) + strwei_ad
    strwei_ad=0._jprb



    DO d = 1, nid
      IF ( mfasis_coefs%lut_axes(di(d))%dim_type .EQ. mfasis_dim_scaangle ) THEN
!       refl_clr_ad = refl_clr_ad+mfasis_refl(0,i)%refl_lin_coef(d)*albedo_ad
        albedo_ad = albedo_ad + mfasis_refl(0,i)%refl_lin_coef(d)* refl_clr_ad
      ELSE
        iw_ad(d)    = iw_ad(d)   +mfasis_refl(0,i)%refl_lin_coef(d)* refl_clr_ad
      ENDIF
    ENDDO
    refl_clr_ad = 0


    DO d = 1, nid
      IF ( mfasis_coefs%lut_axes(di(d))%dim_type .EQ. mfasis_dim_opdp ) THEN
        iw_ad(d) = 0._jprb
      ENDIF
    ENDDO

!=====================================================================================

    DO cs = 1, nstrms
!==================================================================================
! First compute nonlinear quatities needed in linear computations (for given channel "i" and stream "cs")
!==================================================================================
      ! --------------------------------------------------------------------------
      ! Set stream weight, optical depth and effective diameters
      ! --------------------------------------------------------------------------
      IF ( mfasis_coefs%file_type .EQ. mfasis_cld ) THEN ! clouds
        strwei_cs = ircld%xstr(cs+1,prof) - ircld%xstr(cs,prof)
!       strwei_ad = ircld_ad%xstr(cs+1,prof) - ircld_ad%xstr(cs,prof)
        od   (:,:) = 0._jprb
!       od_ad(:,:) = 0._jprb
        ed   (:,:) = 0._jprb
!       ed_ad(:,:) = 0._jprb
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        od1_int = 0._jprb
        od1_mixed= 0._jprb
        od2_mixed= 0._jprb
        fac_mx(:)= 0._jprb

!       od1_int_ad = 0._jprb
!       od1_mixed_ad= 0._jprb
!       od2_mixed_ad= 0._jprb
!       fac_mx_ad(:)= 0._jprb
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        DO j = 1, nlay
          IF (j > aux%s(prof)%nearestlev_surf - 1) EXIT    ! Layer is entirely below surface pressure so nothing more to do
          IF ( ircld%icldarr(cs,j,prof) .EQ. 1 ) THEN
            od   (1,j) = SUM(trans_scatt_ir%opdpext(1:nwcl_max,j,i)) ! water clouds
!           od_ad(1,j) = SUM(trans_scatt_ir_ad%opdpext(1:nwcl_max,j,i)) ! water clouds
            od   (2,j) = trans_scatt_ir%opdpext(nwcl_max+1,j,i)        ! ice clouds
!           od_ad(2,j) = trans_scatt_ir_ad%opdpext(nwcl_max+1,j,i)        ! ice clouds
            IF (j == aux%s(prof)%nearestlev_surf - 1) THEN
              ! Modify optical depth in partial layer above surface
              od_v1(:,j) = od   (:,j)
!             od_ad(:,j) = od_ad(:,j) * (1 - aux%s(prof)%pfraction_surf)  &
!                              - od_v1(:,j) * aux_ad%s(prof)%pfraction_surf
              od   (:,j) = od   (:,j) * (1 - aux%s(prof)%pfraction_surf)
            ENDIF
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
            od1_int = od1_int + od(1,j)
!           od1_int_ad = od1_int_ad + od_ad(1,j)
            ooo1(j)=od1_int-opts%dev%od1_thresh
!           ooo1_ad=od1_int_ad
            fac_mx(j)=(1._jprb + TANH(ooo1(j)/opts%dev%o_del1))/2._jprb

!           fac_mx_ad(j)= 0._jprb
!           IF(abs(ooo1(j)/opts%dev%o_del1) < 5._jprb) &
!           fac_mx_ad(j)=(ooo1_ad/opts%dev%o_del1) / (COSH(ooo1(j)/opts%dev%o_del1)**2) /2._jprb

            od1_mixed=od1_mixed+fac_mx(j)*od(1,j)
            od2_mixed=od2_mixed+fac_mx(j)*od(2,j)
!           od1_mixed_ad=od1_mixed_ad+fac_mx_ad(j)*od(1,j)+fac_mx(j)*od_ad(1,j)
!           od2_mixed_ad=od2_mixed_ad+fac_mx_ad(j)*od(2,j)+fac_mx(j)*od_ad(2,j)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
          ENDIF
          IF ( coefs%coef_mfasis_cld%clw_scheme == clw_scheme_deff ) THEN ! Deff water clouds scheme
            ed   (1,j) = aux%clw_dg(j,prof)
!           ed_ad(1,j) = aux_ad%clw_dg(j,prof)
          ELSE ! OPAC water cloud parametrization
            ed   (1,j) = SUM(wcl_deff(1:nwcl_max) * trans_scatt_ir%opdpext(1:nwcl_max,j,i))
!           ed_ad(1,j) = SUM(wcl_deff(1:nwcl_max) * trans_scatt_ir_ad%opdpext(1:nwcl_max,j,i))
            IF ( od(1,j) .GT. 0. ) THEN
              ! TODO: take into account optical depth surface correction in eff diameter calculation
              ed   (1,j) =  ed   (1,j)/SUM(trans_scatt_ir%opdpext(1:nwcl_max,j,i))
!             ed_ad(1,j) =  ed_ad(1,j)/SUM(trans_scatt_ir%opdpext(1:nwcl_max,j,i)) - &
!                           ed(1,j)*SUM(trans_scatt_ir_ad%opdpext(1:nwcl_max,j,i))/&
!                           SUM(trans_scatt_ir%opdpext(1:nwcl_max,j,i))
            ELSE
              ed   (1,j) = 0._jprb
!             ed_ad(1,j) = 0._jprb
            ENDIF
          ENDIF
          ed   (2,j) = aux%ice_dg(j,prof)
!         ed_ad(2,j) = aux_ad%ice_dg(j,prof)

        ENDDO
        DO k = 1, npar
          od_eff   (k) = SUM( od   (k,:))
          od_eff_v1(k) = od_eff   (k)
!         od_eff_ad(k) = SUM( od_ad(k,:))
          IF ( od_eff_v1(k) .GT. 0. ) THEN
            ed_eff   (k) = SUM( od(k,:)*ed(k,:))/od_eff_v1(k)
            ed_eff_v1(k) = ed_eff(k)
!           ed_eff_ad(k) = SUM( od_ad(k,:)*ed(k,:) + od(k,:)*ed_ad(k,:))/od_eff_v1(k)   &
!                          - ed_eff_v1(k) * od_eff_ad(k)/od_eff_v1(k)

          ELSE
            ed_eff   (k) = ed_aux(k)
!           ed_eff_ad(k) = 0._jprb
          ENDIF
        ENDDO
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        xxx1= opts%dev%qq_1*od1_mixed-od2_mixed
        xxx2= opts%dev%qq_2*od_eff(1)-od_eff(2)
!       xxx1_ad= opts%dev%qq_1*od1_mixed_ad-od2_mixed_ad
!       xxx2_ad= opts%dev%qq_2*od_eff_ad(1)-od_eff_ad(2)
        fac_sw= (1._jprb + TANH(xxx1/opts%dev%x_del1))*(1._jprb + TANH(xxx2/opts%dev%x_del2))/4._jprb

!       fac_sw_ad=0._jprb
!       IF(abs(xxx1/opts%dev%x_del1) < 5._jprb) &
!         fac_sw_ad= (xxx1_ad/opts%dev%x_del1) / &
!                    (COSH(xxx1/opts%dev%x_del1)**2) *(1._jprb + TANH(xxx2/opts%dev%x_del2)) /4._jprb
!       IF(abs(xxx2/opts%dev%x_del2) < 5._jprb) &
!         fac_sw_ad= fac_sw_ad + (xxx2_ad/opts%dev%x_del2) / &
!                    (COSH(xxx2/opts%dev%x_del2)**2) *(1._jprb + TANH(xxx1/opts%dev%x_del1)) /4._jprb


        od_eff(1) = od_eff(1) + fac_sw*od2_mixed
        od_eff(2) = od_eff(2) - fac_sw*od2_mixed
!       od_eff_ad(1) = od_eff_ad(1) + fac_sw_ad*od2_mixed + fac_sw*od2_mixed_ad
!       od_eff_ad(2) = od_eff_ad(2) -(fac_sw_ad*od2_mixed + fac_sw*od2_mixed_ad)

        IF(od_eff(2)>0.) THEN
          ed_eff(2) = SUM( od(2,:)*ed(2,:)*(1._jprb-fac_sw*fac_mx(:)))/od_eff(2)
!         ed_eff_ad(2) = SUM( &
!                              (od_ad(2,:)*ed(2,:) + od(2,:)*ed_ad(2,:))*(1._jprb-fac_sw*fac_mx(:)) - &
!                               od(2,:)*ed(2,:)* ( fac_sw_ad*fac_mx(:)+fac_sw*fac_mx_ad(:) )          &
!                             ) /od_eff(2)   &
!                          - ed_eff(2) * od_eff_ad(2)/od_eff(2)
        ELSE
          ed_eff   (2) = ed_aux(2)
!         ed_eff_ad(2) = 0._jprb
        ENDIF



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      ELSE  ! aerosols
        strwei_cs = 1._jprb     ! single stream for aerosols
!       strwei_ad = 0._jprb     ! single stream for aerosols
        od   (:,:)= 0._jprb
!       od_ad(:,:)= 0._jprb
        DO k = 1, npar
          par = mfasis_coefs%aer_types(k)
          DO j = 1, nlay
            IF (j > aux%s(prof)%nearestlev_surf - 1) EXIT    ! Layer is entirely below surface pressure so nothing more to do
            od(k,j)    = trans_scatt_ir%opdpext(par,j,i)
!           od_ad(k,j) = trans_scatt_ir_ad%opdpext(par,j,i)
            IF (j == aux%s(prof)%nearestlev_surf - 1) THEN
              ! Modify optical depth in partial layer above surface
              od_v1(:,j) = od   (:,j)
!             od_ad(k,j) = od_ad(k,j) * (1 - aux%s(prof)%pfraction_surf)  &
!                         - od_v1(k,j) * aux_ad%s(prof)%pfraction_surf
              od   (k,j) = od   (k,j) * (1 - aux%s(prof)%pfraction_surf)
            ENDIF
          ENDDO
          od_eff   (k) = SUM(od   (k,:))
!         od_eff_ad(k) = SUM(od_ad(k,:))
        ENDDO
      ENDIF
      ! --------------------------------------------------------------------------
      ! Fill into the interpolation point array the rest of dimensions
      ! --------------------------------------------------------------------------
      j = 1
      k = 1
      DO d = 1, nid
        SELECT CASE (mfasis_coefs%lut_axes(di(d))%dim_type)
        CASE (mfasis_dim_opdp)
          ip   (d) = od_eff   (j)
!         ip_ad(d) = od_eff_ad(j)
          j = j + 1
        CASE (mfasis_dim_effdia)
          ! Check IF stored as radius or diamater in LUT
          IF (mfasis_coefs%lut_axes(di(d))%name(1:1) .EQ. "R") THEN
            ip(d) = ed_eff(k)/2
!           ip_ad(d) = ed_eff_ad(k)/2
          ELSE
            ip(d) = ed_eff(k)
!           ip_ad(d) = ed_eff_ad(k)
          ENDIF
          k = k + 1
        END SELECT
      ENDDO

!==================================================================================
! Start linear computations for channel "i" and stream "cs"
!==================================================================================
!     refl_ad = refl_ad + rflstr_ad*strwei_cs + mfasis_refl(cs,i)%refl*strwei_ad
      rflstr_ad = rflstr_ad+  strwei_cs        * refl_ad
      strwei_ad = strwei_ad+   mfasis_refl(cs,i)%refl * refl_ad                  ! rflstr !NNNNNNNNN


      DO d = 1, nid
        IF ( mfasis_coefs%lut_axes(di(d))%dim_type .EQ. mfasis_dim_scaangle ) THEN
!         rflstr_ad = rflstr_ad+mfasis_refl(cs,i)%refl_lin_coef(d)*albedo_ad
          albedo_ad = albedo_ad +mfasis_refl(cs,i)%refl_lin_coef(d)* rflstr_ad
        ELSE
!         rflstr_ad = rflstr_ad+mfasis_refl(cs,i)%refl_lin_coef(d)*iw_ad(d)
          iw_ad(d)  = iw_ad(d) +mfasis_refl(cs,i)%refl_lin_coef(d)* rflstr_ad
        ENDIF
      ENDDO
      rflstr_ad = 0


      DO d = 1, nid
        IF (mfasis_coefs%lut_axes(di(d))%dim_type .EQ. mfasis_dim_scaangle) CYCLE ! Skip scattering angle (already done)
        axis = mfasis_coefs%lut_axes(di(d))

        IF ( ip(d) .GE. axis%values(axis%nvalues) ) THEN
          iw_ad(d)  = 0._jprb
        ELSE
!         DO j = 1, axis%nvalues - 1
          DO j = axis%nvalues - 1 ,1 ,-1
            IF ( ip(d) .GT. axis%values(j) ) THEN
              IF ( mfasis_coefs%lut_axes(di(d))%dim_type .EQ. mfasis_dim_opdp .AND. j .NE. 1 ) THEN
                ! optical depths ( lin. int. in log)
!               iw_ad(d) =  ip_ad(d)/ip(d)                   &
!                 / ( LOG(axis%values(j+1)) - LOG(axis%values(j)) )
                ip_ad(d) = ip_ad(d) + iw_ad(d)/ip(d)                   &
                  / ( LOG(axis%values(j+1)) - LOG(axis%values(j)) )
                iw_ad(d) = 0
              ELSE            ! effective radii and alpha, linear interpolation
!               iw_ad(d) =  ip_ad(d)               &
!                 / ( axis%values(j+1) - axis%values(j) )
                ip_ad(d) = ip_ad(d) + iw_ad(d)               &
                  / ( axis%values(j+1) - axis%values(j) )
                iw_ad(d) = 0
              ENDIF
            ENDIF
          ENDDO
        ENDIF
        iw_ad(d)  = 0._jprb
      ENDDO

!     ! --------------------------------------------------------------------------
!     ! Fill into the interpolation point array the rest of dimensions
!     ! --------------------------------------------------------------------------

      j = 1
      k = 1
      DO d = 1, nid
        SELECT CASE (mfasis_coefs%lut_axes(di(d))%dim_type)
        CASE (mfasis_dim_opdp)
!         ip_ad(d) = od_eff_ad(j)
          od_eff_ad(j)= od_eff_ad(j) + ip_ad(d)
          ip_ad(d) = 0
          j = j + 1
        CASE (mfasis_dim_effdia)
          ! Check if stored as radius or diamater in LUT
          IF (mfasis_coefs%lut_axes(di(d))%name(1:1) .EQ. "R") THEN
            ed_eff_ad(k) = ed_eff_ad(k) + ip_ad(d)/2
          ELSE
            ed_eff_ad(k) = ed_eff_ad(k) + ip_ad(d)
          ENDIF
          ip_ad(d) = 0
          k = k + 1
        END SELECT
      ENDDO

!==================================================================================
!  IF CLOUD ELSE AEROSOL
!==================================================================================
      ! --------------------------------------------------------------------------
      ! Set stream weight, optical depth and effective diameters
      ! --------------------------------------------------------------------------
      IF ( mfasis_coefs%file_type .EQ. mfasis_cld ) THEN ! clouds
!=========================!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!1
        IF( od_eff(2) >0.) THEN
!         ed_eff_ad(2) = SUM( &
!                              (od_ad(2,:)*ed(2,:) + od(2,:)*ed_ad(2,:))*(1._jprb-fac_sw*fac_mx(:)) - &
!                               od(2,:)*ed(2,:)* ( fac_sw_ad*fac_mx(:)+fac_sw*fac_mx_ad(:) )          &
!                             ) /od_eff(2)   &
!                          - ed_eff(2) * od_eff_ad(2)/od_eff(2)
          od_ad(2,:)   = od_ad(2,:)    + ed(2,:)*(1._jprb-fac_sw*fac_mx(:))*ed_eff_ad(2)/od_eff(2)
          ed_ad(2,:)   = ed_ad(2,:)    + od(2,:)*(1._jprb-fac_sw*fac_mx(:))*ed_eff_ad(2)/od_eff(2)
          fac_sw_ad    = fac_sw_ad     - sum(od(2,:)*ed(2,:)*fac_mx(:))    *ed_eff_ad(2)/od_eff(2)
          fac_mx_ad(:) = fac_mx_ad(:)  - od(2,:)*ed(2,:)*fac_sw            *ed_eff_ad(2)/od_eff(2)
          od_eff_ad(2) = od_eff_ad(2)  - ed_eff(2)                      *ed_eff_ad(2)/od_eff(2)
          ed_eff_ad(2) = 0
        ELSE
          ed_eff_ad(2) = 0._jprb
        ENDIF

!       od_eff_ad(2) = od_eff_ad(2) -(fac_sw_ad*od2_mixed + fac_sw*od2_mixed_ad)
        fac_sw_ad    = fac_sw_ad    - od2_mixed * od_eff_ad(2)
        od2_mixed_ad = od2_mixed_ad - fac_sw    * od_eff_ad(2)
!       od_eff_ad(1) = od_eff_ad(1) + fac_sw_ad*od2_mixed + fac_sw*od2_mixed_ad
        fac_sw_ad    = fac_sw_ad    + od2_mixed * od_eff_ad(1)
        od2_mixed_ad = od2_mixed_ad + fac_sw    * od_eff_ad(1)

!       IF(abs(xxx2/opts%dev%x_del2) < 5._jprb) &
!         fac_sw_ad= fac_sw_ad + (xxx2_ad/opts%dev%x_del2) / &
!                    (COSH(xxx2/opts%dev%x_del2)**2) *(1._jprb + TANH(xxx1/opts%dev%x_del1)) /4._jprb
        IF(abs(xxx2/opts%dev%x_del2) < 5._jprb) THEN
          xxx2_ad   = xxx2_ad + (fac_sw_ad/opts%dev%x_del2) / &
                      (COSH(xxx2/opts%dev%x_del2)**2) *(1._jprb + TANH(xxx1/opts%dev%x_del1)) /4._jprb
        ENDIF

!       IF(abs(xxx1/opts%dev%x_del1) < 5._jprb) &
!         fac_sw_ad= (xxx1_ad/opts%dev%x_del1) / (COSH(xxx1/opts%dev%x_del1)**2) *(1._jprb + TANH(xxx2/opts%dev%x_del2)) /4._jprb
        IF(abs( xxx1/opts%dev%x_del1 ) < 5._jprb) THEN
          xxx1_ad   = xxx1_ad + (fac_sw_ad/opts%dev%x_del1) / &
                      (COSH(xxx1/opts%dev%x_del1)**2) *(1._jprb + TANH(xxx2/opts%dev%x_del2)) /4._jprb
          fac_sw_ad = 0
        ENDIF

        fac_sw_ad=0._jprb

!       xxx2_ad= opts%dev%qq_2*od_eff_ad(1)-od_eff_ad(2)
        od_eff_ad(1) = od_eff_ad(1) + opts%dev%qq_2*xxx2_ad
        od_eff_ad(2) = od_eff_ad(2) -      xxx2_ad
        xxx2_ad      = 0

!       xxx1_ad= opts%dev%qq_1*od1_mixed_ad-od2_mixed_ad
        od1_mixed_ad = od1_mixed_ad + opts%dev%qq_1*xxx1_ad
        od2_mixed_ad = od2_mixed_ad -      xxx1_ad
        xxx1_ad= 0
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

        DO k = 1, npar
          IF ( od_eff_v1(k) .GT. 0. ) THEN
!           ed_eff_ad(k) = SUM( od_ad(k,:)*ed(k,:) + od(k,:)*ed_ad(k,:))/od_eff_v1(k)   &
!                          - ed_eff_v1(k) * od_eff_ad(k)/od_eff_v1(k)
            od_ad(k,:) = od_ad(k,:) + ed(k,:) * ed_eff_ad(k) /od_eff_v1(k)
            ed_ad(k,:) = ed_ad(k,:) + od(k,:) * ed_eff_ad(k) /od_eff_v1(k)
            od_eff_ad(k) = od_eff_ad(k) - ed_eff_v1(k) * ed_eff_ad(k) /od_eff_v1(k)
            ed_eff_ad(k) = 0._jprb
          ELSE
            ed_eff_ad(k) = 0._jprb
          ENDIF
!         od_eff_ad(k) = SUM( od_ad(k,:))
          od_ad(k,:) = od_ad(k,:) + od_eff_ad(k)
          od_eff_ad(k) = 0._jprb
        ENDDO


        od1_int_ad = 0
        DO j = nlay, 1, -1
          IF (j > aux%s(prof)%nearestlev_surf - 1) CYCLE    ! Layer is entirely below surface pressure so nothing more to do
!         ed_ad(2,j) = aux_ad%ice_dg(j,prof)
          aux_ad%ice_dg(j,prof) = aux_ad%ice_dg(j,prof) + ed_ad(2,j)
          ed_ad(2,j) = 0._jprb
          IF ( coefs%coef_mfasis_cld%clw_scheme == clw_scheme_deff ) THEN ! Deff water clouds scheme
!           ed_ad(1,j) = aux_ad%clw_dg(j,prof)
            aux_ad%clw_dg(j,prof) = aux_ad%clw_dg(j,prof) + ed_ad(1,j)
            ed_ad(1,j) = 0._jprb
          ELSE ! OPAC water cloud parametrization
            ! TODO: take into account optical depth surface correction in eff diameter calculation
            IF ( od(1,j) .GT. 0._jprb ) THEN
!             ed_ad(1,j) =  ed_ad(1,j)/SUM(trans_scatt_ir%opdpext(1:nwcl_max,j,i)) - &
!                           ed(1,j)*SUM(trans_scatt_ir_ad%opdpext(1:nwcl_max,j,i))/&
!                           SUM(trans_scatt_ir%opdpext(1:nwcl_max,j,i))
              trans_scatt_ir_ad%opdpext(1:nwcl_max,j,i)=trans_scatt_ir_ad%opdpext(1:nwcl_max,j,i) &
                                                        -ed(1,j)/SUM(trans_scatt_ir%opdpext(1:nwcl_max,j,i)) &
                                                         * ed_ad(1,j)
              ed_ad(1,j) =  ed_ad(1,j)/SUM(trans_scatt_ir%opdpext(1:nwcl_max,j,i))
            ELSE
              ed_ad(1,j) = 0._jprb
            ENDIF
!           ed_ad(1,j) = SUM(wcl_deff(1:nwcl_max) * trans_scatt_ir_ad%opdpext(1:nwcl_max,j,i))
            trans_scatt_ir_ad%opdpext(1:nwcl_max,j,i) = trans_scatt_ir_ad%opdpext(1:nwcl_max,j,i) +  &
                                                        wcl_deff(1:nwcl_max) * ed_ad(1,j)
            ed_ad(1,j) = 0
          ENDIF

          IF ( ircld%icldarr(cs,j,prof) .EQ. 1 ) THEN
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!             od2_mixed_ad=od2_mixed_ad+fac_mx_ad(j)*od(2,j)+fac_mx(j)*od_ad(2,j)
              fac_mx_ad(j) = fac_mx_ad(j) + od2_mixed_ad * od(2,j)
              od_ad(2,j)   = od_ad(2,j)   + od2_mixed_ad * fac_mx(j)

!             od1_mixed_ad=od1_mixed_ad+fac_mx_ad(j)*od(1,j)+fac_mx(j)*od_ad(1,j)
              fac_mx_ad(j) = fac_mx_ad(j) + od1_mixed_ad * od(1,j)
              od_ad(1,j)   = od_ad(1,j)   + od1_mixed_ad * fac_mx(j)

!           IF(abs(ooo1(j)/opts%dev%o_del1) < 5._jprb) &
!           fac_mx_ad(j)=(ooo1_ad/opts%dev%o_del1) / (COSH(ooo1(j)/opts%dev%o_del1)**2) /2._jprb
            IF(abs( ooo1(j)/opts%dev%o_del1 ) < 5._jprb) THEN
               ooo1_ad = ooo1_ad + fac_mx_ad(j)/opts%dev%o_del1 / (COSH(ooo1(j)/opts%dev%o_del1)**2) /2._jprb
               fac_mx_ad(j)= 0
            ENDIF

            fac_mx_ad(j)= 0._jprb

!           ooo1_ad=od1_int_ad
            od1_int_ad = od1_int_ad + ooo1_ad
            ooo1_ad    = 0

!           od1_int_ad = od1_int_ad + od_ad(1,j)
            od_ad(1,j) = od_ad(1,j) + od1_int_ad
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
            IF (j == aux%s(prof)%nearestlev_surf - 1) THEN
              ! Modify optical depth in partial layer above surface
!             od_ad(:,j) = od_ad(:,j) * (1 - aux%s(prof)%pfraction_surf)  &
!                              - od_v1(:,j) * aux_ad%s(prof)%pfraction_surf
              aux_ad%s(prof)%pfraction_surf = aux_ad%s(prof)%pfraction_surf &
                               - SUM(od_v1(:,j) * od_ad(:,j))
              od_ad(:,j) = od_ad(:,j) * (1 - aux%s(prof)%pfraction_surf)
            ENDIF
!           od_ad(1,j) = SUM(trans_scatt_ir_ad%opdpext(1:5,j,i)) ! water clouds
            trans_scatt_ir_ad%opdpext(1:nwcl_max,j,i) = trans_scatt_ir_ad%opdpext(1:nwcl_max,j,i) + od_ad(1,j)
            od_ad(1,j) = 0
!           od_ad(2,j) = trans_scatt_ir_ad%opdpext(6,j,i)        ! ice clouds
            trans_scatt_ir_ad%opdpext(nwcl_max+1,j,i) = trans_scatt_ir_ad%opdpext(nwcl_max+1,j,i) + od_ad(2,j)
            od_ad(2,j) = 0
          ENDIF
        ENDDO

!!!!!!!!!!!!!!!!!!!!!!!
        od1_int_ad  = 0._jprb
        od1_mixed_ad= 0._jprb
        od2_mixed_ad= 0._jprb
        fac_mx_ad(:)= 0._jprb
!!!!!!!!!!!!!!!!!!!!!!!
        od_ad(:,:) = 0._jprb
        ed_ad(:,:) = 0._jprb
!       strwei_ad = ircld_ad%xstr(cs+1,prof) - ircld_ad%xstr(cs,prof)
        ircld_ad%xstr(cs+1,prof) = ircld_ad%xstr(cs+1,prof) + strwei_ad
        ircld_ad%xstr(cs,prof)   = ircld_ad%xstr(cs,prof)   - strwei_ad
        strwei_ad = 0
      ELSE  ! Aerosols (not yet tested!!!!)
      ! TODO: Test aerosol functionality

        strwei_cs = 1._jprb     ! single stream for aerosols
        od   (:,:)= 0._jprb
        DO k = 1, npar
          par = mfasis_coefs%aer_types(k)
!         od_eff_ad(k) = SUM(od_ad(k,:))
          od_ad(k,:) = od_ad(k,:) + od_eff_ad(k)
          od_eff_ad(k) = 0
          DO j = 1, nlay
            IF (j > aux%s(prof)%nearestlev_surf - 1) EXIT    ! Layer is entirely below surface pressure so nothing more to do
            od(k,j)    = trans_scatt_ir%opdpext(par,j,i)
            IF (j == aux%s(prof)%nearestlev_surf - 1) THEN
              ! Modify optical depth in partial layer above surface
!             od_ad(k,j) = od_ad(k,j) * (1 - aux%s(prof)%pfraction_surf)  &
!                        - od_v1(k,j) * aux_ad%s(prof)%pfraction_surf
              aux_ad%s(prof)%pfraction_surf = aux_ad%s(prof)%pfraction_surf - od_v1(k,j) * od_ad(k,j)
              od_ad(k,j) = (1 - aux%s(prof)%pfraction_surf) * od_ad(k,j)
            ENDIF
!           od_ad(k,j) = trans_scatt_ir_ad%opdpext(par,j,i)
            trans_scatt_ir_ad%opdpext(par,j,i) = trans_scatt_ir_ad%opdpext(par,j,i) + od_ad(k,j)
            od_ad(k,j) = 0
          ENDDO
        ENDDO
        strwei_ad = 0._jprb     ! single stream for aerosols A1
        od_ad(:,:)= 0._jprb     !                            A2
      ENDIF !ENDIF CLOUD ELSE AEROSOL

    ENDDO ! END of cloud stream loop

!====================================================================================
    refl_ad     = 0._jprb
    refl_clr_ad = 0._jprb

    iw_ad(:)  = 0._jprb  ! interpolation weight
    ip_ad(:) =  0._jprb         ! interpolation point
    ! --------------------------------------------------------------------------
    ! Albedo
    ! --------------------------------------------------------------------------
!   albedo = MIN(reflectance(i)%refl_out * pi, 1._jprb)
    albedo   =     reflectance(i)%refl_out * pi
    IF (albedo < 1._jprb) THEN
!     albedo_ad= reflectance_ad(i)%refl_out * pi
      reflectance_ad(i)%refl_out = reflectance_ad(i)%refl_out + albedo_ad * pi
      albedo_ad= 0
    ELSE
      albedo_ad= 0._jprb
    ENDIF
  ENDDO ! end of channel loop

  ! --------------------------------------------------------------------------
  ! Tidy up
  ! --------------------------------------------------------------------------

  DEALLOCATE( ip, ip_ad, iw_ad, di, od, od_ad, od_eff, od_eff_ad, od_v1, od_eff_v1, STAT=err)
  THROWM(err.NE.0,"Deallocation of memory for rttov_mfasis_ad failed")
  IF (mfasis_coefs%file_type .EQ. mfasis_cld) THEN
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    DEALLOCATE(ed, ed_ad, ed_eff, ed_eff_v1, ed_aux, ed_eff_ad, fac_mx, fac_mx_ad, STAT=err)
    THROWM(err.NE.0,"Deallocation of memory for rttov_mfasis_ad failed")
  ENDIF


  CATCH

END SUBROUTINE rttov_mfasis_ad
